import logging

from apistar import ASyncApp, App

from .bases.service import Service
from .bases.components import Component, SettingsComponent
from .helper import load_packages, routing, print_routing

__all__ = ["Application"]

logger = logging.getLogger("star_builder.app")


def application(template_dir=None,
                static_dir=None,
                packages=None,
                schema_url='/schema/',
                docs_url='/docs/',
                static_url='/static/',
                components=None,
                event_hooks=None,
                settings_path="settings",
                debug=True,
                async=True):
    """
       参数指定选择使用异步app还是同步app
       可以动态发现当前项目根目录下所有routes的service中的handler
    """
    load_packages(".")
    include = routing(Service, None)
    SettingsComponent.register_path(settings_path)
    loaded, unloaded = [], []

    def find_children(components):
        children = []
        for component in components:
            children.append(component)
            children.extend(find_children(component.children))
        return children

    for child in find_children(Component.children):
        if child._instance:
            loaded.append(child._instance)
        else:
            unloaded.append(child)

    components = [c() for c in sorted(unloaded, key=lambda x: -x.order)] + \
                 (components or []) + loaded

    if include:
        routes = [include]
        print_routing(routes, callback=logger.debug)
    else:
        logger.info("Noting to route. ")
        routes = []

    app = (ASyncApp if async else App)(
        routes,
        template_dir=template_dir,
        static_dir=static_dir,
        packages=packages,
        schema_url=schema_url,
        docs_url=docs_url,
        static_url=static_url,
        components=components,
        event_hooks=event_hooks)

    app.debug = debug
    return app


Application = application



import os
import sys

from argparse import ArgumentParser
from jinja2 import PackageLoader, Environment, FileSystemLoader

from ..helper import ArgparseHelper
from .tasks import *


class Command(object):
    """
    项目构建工具
    """
    def __init__(self, tasks):
        self.tasks = tasks
        self.args = self.parse_args()
        self.templates_path = self.args.templates
        self.task = tasks[self.args.task.lower()]()

    def create(self):
        if self.templates_path:
            env = Environment(loader=FileSystemLoader(self.templates_path))
        else:
            env = Environment(loader=PackageLoader('star_builder', 'templates'))
        self.task.create(env, **vars(self.args))

    def parse_args(self):
        base_parser = ArgumentParser(
            description=self.__class__.__doc__, add_help=False)
        base_parser.add_argument("-t", "--templates", help="模板路径.")

        parser = ArgumentParser(description="Apistar项目构建工具", add_help=False)
        parser.add_argument('-h', '--help', action=ArgparseHelper,
                            help='显示帮助信息并退出. ')
        sub_parsers = parser.add_subparsers(dest="task", help="创建模块类型. ")

        for name, task in self.tasks.items():
            sub_parser = sub_parsers.add_parser(
                name.lower(), parents=[base_parser], help=task.__doc__)
            task.enrich_parser(sub_parser)
        # 当不提供参数时，len(sys.argv) == 1, 不会打印帮助，需要手动打印
        # add_subparsers无法指定required=True, 这是argparse的bug。
        if len(sys.argv) < 2:
            parser.print_help()
            exit(1)
        return parser.parse_args()


def find_tasks():
    sys.path.insert(0, os.getcwd())
    tasks = dict()

    try:
        model = __import__("tasks")
        for k in dir(model):
            cls = getattr(model, k)
            if cls is not Task and isinstance(cls, type) and issubclass(cls, Task):
                tasks[k.lower()] = cls
    except ImportError:
        pass

    for name, prop in globals().items():
        if prop is not Task and isinstance(prop, type) and issubclass(prop, Task):
            tasks[name.lower()] = prop
    return tasks


def main():
    Command(find_tasks()).create()


if __name__ == "__main__":
    main()



from threading import RLock
from apistar import Component


class Meta(type):
    """
    元类的主要作用是建立继承树，以便发现所有Service/Component。
    """
    lock = RLock()

    def __new__(mcs, class_name, bases, props):
        props["_instance"] = None
        props["children"] = []

        if len(bases) > 1:
            raise RuntimeError(f"{class_name} unsupprt multi inherit!")

        cls = super().__new__(mcs, class_name, bases, props)
        if not bases or bases[0] in [object, Component]:
            return cls

        bases[0].children.append(cls)
        return cls


class ComponentMeta(Meta):
    """
    单例，修改resolve return annotation
    """
    def __new__(mcs, class_name, bases, props):
        props["_instance"] = None
        return super().__new__(mcs, class_name, bases, props)

    def __call__(cls, *args, **kwargs):
        with cls.lock:
            cls._instance = cls._instance or super().__call__(*args, **kwargs)
            return cls._instance



import inspect

from toolkit.frozen import FrozenSettings
from toolkit.settings import SettingsLoader
from apistar import Route, Component as _Component

from . import ComponentMeta
from .service import Service


class Component(_Component, metaclass=ComponentMeta):
    order = 1000

    def resolve(self, *args, **kwargs):
        raise NotImplementedError()


class ServiceComponent(Component):
    """
    注入Service
    """
    order = 1

    def resolve(self, route: Route) -> Service:
        return route.service


class SettingsComponent(Component):
    """
    注入Settings
    """
    settings_path = None
    order = 0

    def __init__(self):
        self.settings = SettingsLoader().load(self.settings_path or "settings")

    def resolve(self) -> FrozenSettings:
        return self.settings

    @classmethod
    def register_path(cls, settings_path):
        cls.settings_path = settings_path



import os
import glob

from apistar import Include
from argparse import Action, _SubParsersAction


def print_routing(routes, callback=print, parent=""):
    for route in routes:
        if isinstance(route, Include):
            print_routing(route.routes, callback ,parent + route.url)
        else:
            callback(
                f"Route method: {route.method}, "
                f"url: {parent + route.url} to {route.name}.")


def load_packages(current_path):
    """
    加载当前路径下的所有package，使得其中的Service子类得以激活
    加载一个包时，如果包下面有子包，只需要导入子包，父包也会一起
    被加载。项目约定service子类必须定义在包中(__init__.py)。
    所以只考虑加载所有包，不考虑加载其它模块。
    :param current_path:
    :return:
    """
    files = glob.glob(os.path.join(current_path, "*"))
    find_dir = False

    for file in files:
        if os.path.isdir(file):
            find_dir = True
            if not load_packages(file):
                __import__(file.replace("/", ".").strip("."))

    return find_dir


def routing(service, parent_prefix):
    """
    获取当前Service下所有route及其子Service组成的Include
    :param service:
    :param parent_prefix:
    :return:
    """
    if not hasattr(service, "prefix") or service.prefix == parent_prefix:
        raise RuntimeError(f"{service} is not routed! ")
    routes = []
    instance = service()
    for name in vars(service).keys():
        prop = getattr(service, name)
        if hasattr(prop, "routes"):
            for route in prop.routes:
                route.service = instance
            routes.extend(prop.routes)

    for child_service in service.children:
        child_include = routing(child_service, service.prefix)
        if child_include:
            routes.append(child_include)

    if routes:
        return Include(service.prefix, service.name, routes)


class ArgparseHelper(Action):
    """
        显示格式友好的帮助信息
    """

    def __init__(self,
                 option_strings,
                 dest="",
                 default="",
                 help=None):
        super(ArgparseHelper, self).__init__(
            option_strings=option_strings,
            dest=dest,
            default=default,
            nargs=0,
            help=help)

    def __call__(self, parser, namespace, values, option_string=None):
        parser.print_help()
        subparsers_actions = [
            action for action in parser._actions
            if isinstance(action, _SubParsersAction)]
        for subparsers_action in subparsers_actions:
            for choice, subparser in subparsers_action.choices.items():
                print("Command '{}'".format(choice))
                print(subparser.format_usage())

        parser.exit()



import re

from os import makedirs
from os.path import join, exists
from abc import ABC, abstractmethod


__all__ = ["Task", "Project", "Service"]


class Task(ABC):

    @abstractmethod
    def create(self, env, **kwargs):
        pass

    @classmethod
    @abstractmethod
    def enrich_parser(self, sub_parser):
        pass


class Project(Task):
    """
    项目
    """
    def create(self, env, **kwargs):
        task = kwargs.pop("task")
        names = kwargs.pop("name")
        for name in names:
            makedirs(join(name, name), exist_ok=True)
            template = env.get_template(join(task, 'start.py.tmpl'))
            with open(join(name, "start.py"), "w") as f:
                f.write(template.render())
            print(f"{name} 项目已完成创建。")

    @classmethod
    def enrich_parser(cls, sub_parser):
        sub_parser.add_argument("name", nargs=1, help="项目名称")


class Service(Task):
    """
    服务
    """
    def create(self, env, **kwargs):
        task = kwargs.pop("task")
        names = kwargs.pop("name")
        father = None
        if exists("__init__.py"):
            regex = re.compile(r"class\s+(\w*?Service)\(\w*Service\):")
            mth = regex.search(open("__init__.py").read())
            if mth:
                father = mth.group(1)

        for name in names:
            words = re.findall(r"([A-Za-z0-9]+)", name)
            assert words, f"name: {name} invalid!"
            assert words[0][0].isalpha(), f"name: {name} start with number!"
            makedirs(name, exist_ok=True)
            init = env.get_template(join(task, '__init__.py.tmpl'))
            with open(join(name, "__init__.py"), "w") as f:
                f.write(init.render(father=father or "Service", service=name))

        print("、".join(names), "服务模块已完成创建。")

    @classmethod
    def enrich_parser(cls, sub_parser):
        sub_parser.add_argument("name", nargs="+", help="服务模块名称")



# -*- coding:utf-8 -*-
from .build import main
from .app import Application
from .bases.service import Service
from .bases.components import Component
from .route import route, get, post, delete, put, options


__version__ = "0.1.9"
